from typing import Iterator

import requests

from . import __version__
from .model import ISDNRecord, ISDNRoot
from .parser import ISDNJpSitemapXMLParser

ISDN_XML_ENDPOINT = "https://isdn.jp/xml/{isdn}"
ISDN_IMAGE_ENDPOINT = "https://isdn.jp/images/thumbs/{isdn}.png"
ISDN_SITEMAP = "https://isdn.jp/sitemap.xml"


class ISDNClient:
    def __init__(
        self,
        xml_endpoint_url: str = ISDN_XML_ENDPOINT,
        image_endpoint_url: str = ISDN_IMAGE_ENDPOINT,
        sitemap_url: str = ISDN_SITEMAP,
    ):
        self.xml_endpoint_url = xml_endpoint_url
        self.image_endpoint_url = image_endpoint_url
        self.sitemap_url = sitemap_url
        self.s = requests.Session()
        self.set_user_agent(f"isdn-python/{__version__}")

    def set_user_agent(self, user_agent: str):
        self.s.headers.update({"user-agent": user_agent})

    @staticmethod
    def normalize_isdn(isdn: str) -> str:
        return isdn.replace("-", "").strip()

    def _get(self, isdn: str, endpoint_url: str) -> requests.Response:
        r = self.s.get(endpoint_url.format(isdn=self.normalize_isdn(isdn)))
        r.raise_for_status()
        return r

    def get(self, isdn: str) -> ISDNRecord:
        r = self._get(isdn, self.xml_endpoint_url)
        return ISDNRoot.from_xml_first(r.content)

    def get_raw(self, isdn: str) -> bytes:
        r = self._get(isdn, self.xml_endpoint_url)
        return r.content

    def get_image(self, isdn: str) -> bytes:
        r = self._get(isdn, self.image_endpoint_url)
        return r.content

    def _list(self) -> requests.Response:
        r = self.s.get(self.sitemap_url, stream=True)
        r.raise_for_status()
        return r

    def list(self) -> Iterator[str]:
        r = self._list()
        return ISDNJpSitemapXMLParser.parse_list(r.raw)



import re
from typing import IO, Iterator

from lxml import etree

namespaces = {"sitemap": "http://www.sitemaps.org/schemas/sitemap/0.9"}


class ISDNJpSitemapXMLParser:
    @staticmethod
    def parse_list(file: str | IO) -> Iterator[str]:
        for event, elm in etree.iterparse(
            file, events=("end",), tag=[f"{{{namespaces['sitemap']}}}loc"], remove_blank_text=True
        ):
            m = re.match(r"https://isdn.jp/(\d{13})", elm.text)
            if not m:
                continue
            yield m.group(1)



import re
from datetime import date
from decimal import Decimal
from typing import Any

from pydantic import BaseModel, Field, HttpUrl, model_validator
from pydantic_xml import BaseXmlModel, element

from . import InvalidIsdnError

NSMAP = {"": "https://isdn.jp/schemas/0.1"}


class ISDN(BaseModel):
    code: str
    prefix: str | None = None
    group: str | None = None
    registrant: str | None = None
    publication: str | None = None
    check_digit: str | None = None

    def __init__(
        self,
        code: str,
        prefix: str | None = None,
        group: str | None = None,
        registrant: str | None = None,
        publication: str | None = None,
        check_digit: str | None = None,
    ) -> None:
        super().__init__(
            code=code,
            prefix=prefix,
            group=group,
            registrant=registrant,
            publication=publication,
            check_digit=check_digit,
        )

    @model_validator(mode="before")
    @classmethod
    def validate_code(cls, values: Any) -> Any:
        code = str(values["code"])

        if code.startswith("ISDN") and "-" in code:
            parts = code.lstrip("ISDN").split("-")
            code = "".join(parts)
            if len(parts) != 5:
                raise InvalidIsdnError("ISDN must have 5 parts")

            values.update(
                {
                    "prefix": parts[0],
                    "group": parts[1],
                    "registrant": parts[2],
                    "publication": parts[3],
                    "check_digit": parts[4],
                }
            )
        else:
            code = cls.normalize(code)

        arg_parts = [values.get(k) for k in ["prefix", "group", "registrant", "publication", "check_digit"]]
        if all(arg_parts) and code != "".join(arg_parts):
            raise ValueError(f"ISDNs of arguments do not match: {code} != {''.join(arg_parts)}")

        return values | {"code": code}

    @property
    def parts(self) -> list[str | None]:
        self.code = self.normalize(self.code)
        return [self.prefix, self.group, self.registrant, self.publication, self.check_digit]

    @staticmethod
    def normalize(isdn: int | str) -> str:
        return str(isdn).replace("-", "").strip()

    @staticmethod
    def calc_check_digit(isdn: str) -> str:
        isdn = [int(n) for n in isdn]
        cd = 10 - (sum([(n if i % 2 == 0 else n * 3) for i, n in enumerate(isdn[:12])]) % 10)
        return str(cd % 10)

    def to_disp_isdn(self) -> str | None:
        if not all(self.parts):
            return None
        return f"ISDN{self.prefix}-{self.group}-{self.registrant}-{self.publication}-{self.check_digit}"

    def validate(self, raise_error: bool = False) -> bool:
        if not re.fullmatch(r"\d+", self.code):
            if raise_error:
                raise InvalidIsdnError("Contains non-numeric characters")
            else:
                return False
        if len(self.code) != 13:
            if raise_error:
                raise InvalidIsdnError("ISDN must be 13 digits")
            else:
                return False
        if not (self.code.startswith("278") or self.code.startswith("279")):
            if raise_error:
                raise InvalidIsdnError("ISDN prefix must be 278 or 279")
            else:
                return False
        if self.calc_check_digit(self.code) != self.code[12]:
            if raise_error:
                raise InvalidIsdnError("Invalid check digit")
            else:
                return False

        # Validate parts
        if self.group and not (1 <= len(self.group) <= 5):
            if raise_error:
                raise InvalidIsdnError("Group part must be 1 to 5 digits")
            else:
                return False
        if self.registrant and not (1 <= len(self.registrant) <= 7):
            if raise_error:
                raise InvalidIsdnError("Publisher part must be 1 to 7 digits")
            else:
                return False
        if self.publication and not (1 <= len(self.publication) <= 2):
            if raise_error:
                raise InvalidIsdnError("Publication part must be 1 to 2 digits")
            else:
                return False

        return True


class UserOption(BaseXmlModel, tag="useroption", nsmap=NSMAP):
    property: str = element(tag="property", default="")
    value: str = element(tag="value", default="")


class ExternalLink(BaseXmlModel, tag="external-link", nsmap=NSMAP):
    title: str | None = element(tag="title", default=None)
    uri: HttpUrl | str = element(tag="uri")


class ISDNRecord(BaseXmlModel, nsmap=NSMAP):
    disp_isdn: str = element(tag="disp-isdn")
    region: str = element(tag="region")
    class_: str = element(tag="class")
    type: str = element(tag="type")
    rating_gender: str = element(tag="rating_gender")
    rating_age: str = element(tag="rating_age")
    product_name: str | None = element(tag="product-name", default=None)  # 仕様上は必須
    product_yomi: str | None = element(tag="product-yomi", default=None)
    publisher_code: str = element(tag="publisher-code")
    publisher_name: str = element(tag="publisher-name")
    publisher_yomi: str | None = element(tag="publisher-yomi", default=None)
    issue_date: date = element(tag="issue-date")
    genre_code: str | None = element(tag="genre-code", default=None)
    genre_name: str | None = element(tag="genre-name", default=None)
    genre_user: str | None = element(tag="genre-user", default=None)
    c_code: str | None = element(tag="c-code", default=None)
    author: str | None = element(tag="author", default=None)
    shape: str | None = element(tag="shape", default=None)
    contents: str | None = element(tag="contents", default=None)
    price: Decimal | None = element(tag="price", default=None)
    price_unit: str | None = element(tag="price-unit", default=None)
    barcode2: str | None = element(tag="barcode2", default=None)
    product_comment: str | None = element(tag="product-comment", default=None)
    product_style: str | None = element(tag="product-style", default=None)
    product_size: str | None = element(tag="product-size", default=None)
    product_capacity: Decimal | None = element(tag="product-capacity", default=None)
    product_capacity_unit: str | None = element(tag="product-capacity-unit", default=None)
    sample_image_uri: HttpUrl | None = element(tag="sample-image-uri", default=None)
    useroptions: list[UserOption] = Field(default_factory=list)
    external_links: list[ExternalLink] = Field(default_factory=list)

    @property
    def isdn(self) -> ISDN:
        return ISDN(code=self.disp_isdn)


class ISDNRoot(BaseXmlModel, tag="isdn", nsmap=NSMAP):
    records: list[ISDNRecord] = element(tag="item")

    @classmethod
    def from_xml_first(cls, source: str | bytes) -> ISDNRecord:
        isdn_root = cls.from_xml(source)
        return isdn_root.records[0]



import os
import time

import click
from requests.exceptions import HTTPError

from . import __version__
from .client import ISDNClient
from .model import ISDN, ISDNRoot


@click.group()
def cli():
    pass


@cli.command()
def version():
    click.echo(f"isdn/{__version__}")


@cli.command("get", help="Get record from isdn.jp")
@click.argument("isdn")
@click.option("--format", "-f", type=click.Choice(["xml", "dict", "json"]), default="xml")
@click.option("--pretty", "-p", is_flag=True, help="Pretty print output")
def get_isdn(isdn: str, format: str, pretty: bool):
    from pprint import pformat

    c = ISDNClient()
    match format:
        case "xml":
            res = c.get_raw(isdn)
        case "dict":
            res = c.get(isdn).model_dump()
            if pretty:
                res = pformat(res)
        case "json":
            res = c.get(isdn).model_dump_json(ensure_ascii=False, indent=2 if pretty else None)
        case _:
            raise NotImplementedError
    click.echo(res)


@cli.command("list", help="List all ISDNs from isdn.jp")
def list_isdns():
    c = ISDNClient()
    for isdn in c.list():
        click.echo(isdn)


@cli.command(help="Download all xml record files from isdn.jp")
@click.argument("directory", type=click.Path(exists=True, file_okay=False, writable=True))
@click.option("--force", "-f", is_flag=True, help="Overwrite existing files")
@click.option("--stop-on-error", is_flag=True, help="Stop when error occurs during download")
@click.option("--sleep-time", "-s", type=int, default=500, help="Sleep interval for download (millisecond)")
@click.option("--write-image", is_flag=True, help="Write cover image to file")
@click.option(
    "--write-image-path",
    type=click.Path(exists=True, file_okay=False, writable=True),
    help="Directory path to write cover images",
)
def bulk_download(
    directory: str, force: bool, stop_on_error: bool, sleep_time: int, write_image: bool, write_image_path: str
):
    c = ISDNClient()
    with click.progressbar(list(c.list()), show_pos=True) as bar:
        for isdn in bar:
            path = os.path.join(directory, f"{isdn}.xml")
            image_path = os.path.join(write_image_path or directory, f"{isdn}.png")

            if not force and os.path.exists(path) and (not write_image or write_image and os.path.exists(image_path)):
                continue

            record = None
            if not force and os.path.exists(path):
                with open(path, "rb") as f:
                    record = ISDNRoot.from_xml_first(f.read())
                if write_image and not record.sample_image_uri:
                    continue
            else:
                try:
                    res = c.get_raw(isdn)
                    with open(path, "wb") as out:
                        out.write(res)
                    record = ISDNRoot.from_xml_first(res)
                except HTTPError as err:
                    if stop_on_error:
                        raise err
                    else:
                        continue

            if write_image and record.sample_image_uri:
                if not force and os.path.exists(image_path):
                    continue
                img = c.get_image(isdn)
                with open(image_path, "wb") as out:
                    out.write(img)

            time.sleep(sleep_time / 1000)


@cli.command(help="Validate ISDN code")
@click.argument("isdn")
def validate(isdn: str):
    try:
        click.echo(ISDN(isdn).validate(raise_error=True))
    except Exception as e:
        click.echo(e, err=True)
